/*
	ProWesS, PROGS Windowing System

	PWHandleRegion
		See if the pointer is inside one of the children. If this is the case,
		then call the handler routine for that child region.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
	25.11.1995 added support for SSTATUS_PAGESHOW
	04.02.1996 added support for EVENT_TOUCH_EDGE
			   added support for catching software generated events
	12.05.1996 WINPointerLimits now called when dragging
	25.05.1996 proper return of error code caused by type->PWHandler
	29.06.1996 no PWHandleKeypress call on keypress event for hit or do
			   these could generate spurious events when dragging
	14.04.1997 changed event handling to "better" order
	09.10.1997 attempt to fix a problem where you could get both a hit and
			   a space event when pressing space ?? fix tries to solve this
			   by clearing the keypress when handling the keystroke
*/

#include <win.h>

#include "private.h"
#include "PWHandler.h"

extern TimeOut system_dragtimeout;
extern short XWindowSize, YWindowSize;
extern short system_borderwidth;
extern short system_scaleborderwidth;

/* forward declaration */
static Error PWRegionHandler(PWSystem *system, PWRegion *region);

Error PWHandleRegion(PWSystem *system, PWRegion *region)
{
	PWRegion *child;
	char finished=FALSE;
	Error err;
	pix xpos, ypos;

	while (!finished)
	{
		xpos=system->ptr_xpos-system->win_xpos+system->xorg;
		ypos=system->ptr_ypos-system->win_ypos+system->yorg;
		finished=TRUE;

		for (child=region?region->children:system->regions; child; child=child->next)
		{
			if ( xpos >= child->hit_xorg && xpos <= child->hit_xend &&
				 ypos >= child->hit_yorg && ypos <= child->hit_yend )
			{
				err=PWRegionHandler(system,child);
				if (err) return err;
				if (!(system->status&(SSTATUS_BREAKDOWN|SSTATUS_WINCHANGE)))
					finished=FALSE;
				break;
			}
		}
	}
	return ERR_OK;
}

#define EXIT { system->active=previous; return err; }

static PWTypeDef NULLtype={0};
static struct _PWDummyObject NULLowner={&NULLtype,0,0,0};

static void SetLimits(PWSystem *system)
{
	int xb,yb;
	xb=yb=(system_scaleborderwidth+system_borderwidth*2);
	if (system->status&SSTATUS_NOXBORDER) xb=0;
	if (system->status&SSTATUS_NOYBORDER) yb=0;
	xb+=system->win_xpos;
	yb+=system->win_ypos;
	WINPointerLimits(system->window,xb,yb,
					 xb+system->curr_xsiz,yb+system->curr_ysiz);
}

static Error PWRegionHandler(PWSystem *system, PWRegion *region)
{
	Error err=ERR_OK;
	Sprite *sprite=system->subwindows.main.sprite;
	PWObject owner=region->owner;
	PWRegion *previous=system->active;
	WindowEvent event;
	TimeOut timeout;
	char key;
	int following=FALSE;

	system->active=region;

	err=PWHandleRegion(system,region);
	if (err) EXIT;

	if (owner==NULL)				/* make sure I can handle no owner */
		owner=&NULLowner;

	if (owner->type->PWRegionEntry)
	{
		err=owner->type->PWRegionEntry(owner);
		if (err) EXIT;
	}

	while (!(system->status&(SSTATUS_BREAKDOWN|SSTATUS_WINCHANGE)))
	{
		err=PWRedraw(system);
		if (err) EXIT;

		if (following)
		{
			event=EVENT_MOVE|EVENT_KEYSTROKE|EVENT_KEYPRESS|
				  EVENT_OUT_WINDOW|EVENT_TOUCH_EDGE|EVENT_LEVELS;
			timeout=region->wait_event&PW_EVENT_TIMEOUT?region->timeout:FOREVER;
			if (region->dragging) event=EVENT_MOVE|EVENT_KEYUP|EVENT_OUT_WINDOW;
			err=WINPointerRead(system->window,timeout, event,
							   &system->ptr_xpos,&system->ptr_ypos,&system->ptr_info);
			if (err)
			{
				if (err==ERR_NC)
				{
					if (region->wait_event&PW_EVENT_TIMEOUT)
					{
						err=owner->type->PWHandle(owner, PW_EVENT_TIMEOUT);
						if (err) EXIT;
					}
				}
				else
					EXIT;
			}
		}
		following=TRUE;

		err=PWHandleRegion(system,region);
		if (err) EXIT;

		if (system->ptr_info.event & (EVENT_MOVE|EVENT_OUT_WINDOW))
		{
			pix xpos=system->ptr_xpos-system->win_xpos+system->xorg;
			pix ypos=system->ptr_ypos-system->win_ypos+system->yorg;

			/* check if we have moved outside this region */
			if ( xpos < region->hit_xorg || xpos > region->hit_xend ||
				 ypos < region->hit_yorg || ypos > region->hit_yend )
			{
				if (region->dragging)
				{
					if((system->Xdragadjust=region->hit_xorg-xpos)<0)
						if((system->Xdragadjust=region->hit_xend-xpos)>0)
							system->Xdragadjust=0;
					if((system->Ydragadjust=region->hit_yorg-ypos)<0)
						if((system->Ydragadjust=region->hit_yend-ypos)>0)
							system->Ydragadjust=0;
					WINPointerPositionSet(system->window,
							system->ptr_xpos+system->Xdragadjust,
							system->ptr_ypos+system->Ydragadjust);
					if (region->wait_event&PW_EVENT_DRAGADJUST)
					{
						err=owner->type->PWHandle(owner, PW_EVENT_DRAGADJUST);
						if (err) EXIT;
					}
				}
				else
					break;
			}
			else
			{
				if (region->wait_event&PW_EVENT_MOVE)
				{
					err=owner->type->PWHandle(owner,PW_EVENT_MOVE);
					if (err) EXIT;
				}
			}
		}

		if (region->dragging && (system->ptr_info.event&EVENT_KEYUP) )
		{
			region->dragging=FALSE;
			err=owner->type->PWHandle(owner, PW_EVENT_DRAGEND);
			if (err) EXIT;
		}
		else
		if (key=system->ptr_info.keystroke)
		{
			if (key==KEYSTROKE_HIT &&
				region->wait_event&(PW_EVENT_HIT|PW_EVENT_HITDRAG))
			{
				if (region->wait_event&PW_EVENT_HITDRAG &&
					ERR_NC==WINPointerRead(system->window,system_dragtimeout,EVENT_KEYUP,
						&system->ptr_xpos,&system->ptr_ypos,&system->ptr_info))
				{
					region->dragging=TRUE;
					SetLimits(system);
					err=owner->type->PWHandle(owner, PW_EVENT_HITDRAG);
					if (err) EXIT;
				}
				else if (region->wait_event&PW_EVENT_HIT)
				{
					err=owner->type->PWHandle(owner, PW_EVENT_HIT);
					if (err) EXIT;
				}
				system->ptr_info.keypress=0;
			}
			else if (key==KEYSTROKE_DO &&
					 region->wait_event&(PW_EVENT_DO|PW_EVENT_DODRAG))
			{
				if (region->wait_event&PW_EVENT_DODRAG &&
					ERR_NC==WINPointerRead(system->window,system_dragtimeout,EVENT_KEYUP,
						&system->ptr_xpos,&system->ptr_ypos,&system->ptr_info))
				{
					region->dragging=TRUE;
					SetLimits(system);
					err=owner->type->PWHandle(owner, PW_EVENT_DODRAG);
					if (err) EXIT;
				}
				else if (region->wait_event&PW_EVENT_DO)
				{
					err=owner->type->PWHandle(owner, PW_EVENT_DO);
					if (err) EXIT;
				}
				system->ptr_info.keypress=0;
			}
			else
			{
				err=PWHandleKeyPress(system);
				if (err) EXIT;
			}
			system->ptr_info.keystroke=0;
		}
		else if (key=system->ptr_info.keypress && key!=KEYSTROKE_HIT && key!=KEYSTROKE_DO)
		{
			system->ptr_info.keystroke=key;
			err=PWHandleKeyPress(system);
			if (err) EXIT;
		}
		if (system->ptr_info.event & EVENT_TOUCH_EDGE)
			PWScrollTouch(system);
		if (system->ptr_info.event & EVENT_LEVELS)
		{
			err=SoftwareEvent(system);
			if (err) EXIT;
		}
	}
	if (owner->type->PWRegionExit)
		err=owner->type->PWRegionExit(owner);
	system->subwindows.main.sprite=sprite;
	EXIT;
}
